/*
 * Pong: reply to UDP packet and increment a counter
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/socket.h>
#include <linux/in.h>

#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <asm/system.h>
#include <asm/io.h>

#include <rtl.h>
#include <rtl_fifo.h>

#define RTL_SPIN_LOCK SPIN_LOCK_UNLOCKED
#include <rtl_sync.h>
#include "../onetd.h"


MODULE_AUTHOR("Pierre Morel <pmorel@mnis.fr>");
MODULE_DESCRIPTION("Ocera Network ping module example");
MODULE_LICENSE("GPL");

/*
 * Initialize the module, get queues and wait for data
 */

int period = OCS_PERIOD*100 ;

pthread_t	ping_thread;



#define RECEIVE_BUFFER_SIZE	1024
char my_receive_buffer [RECEIVE_BUFFER_SIZE];

#define MSG_LEN	128
char my_message [MSG_LEN];

/*
#define	REM_ADDR	INADDR_LOOPBACK
*/
/* 192.168.254.5
#define	REM_ADDR	0xc0a8fe05
#define	REM_ADDR	INADDR_LOOPBACK
*/
#define	LOC_ADDR	INADDR_ANY
#define MY_PORT		0

#define REM_PORT	(unsigned short) 10061
#define	REM_ADDR	INADDR_LOOPBACK

int sock=-1;
static struct sockaddr_in my_socket;
static struct sockaddr_in rem_socket;
static struct sockaddr_in rem_socket2;
char	*my_name = "ping";

/*
 * PING THREAD
 */

void *ping_rt_thread(void *t){
	int err;
	int count=0;
	int count2=0;
	int len;

	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: ping_rt_thread\n"));
	sprintf(my_message,"RT MESSAGE %d",count);
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: pthread_make_periodic_np\n"));
	pthread_make_periodic_np( pthread_self(), gethrtime(), period );
	rem_socket.sin_family = AF_INET;
	rem_socket.sin_addr.s_addr = htonl(REM_ADDR);
	rem_socket.sin_port = htons(REM_PORT);

	my_socket.sin_family = AF_INET;
	my_socket.sin_addr.s_addr = LOC_ADDR;
	my_socket.sin_port = htons(MY_PORT);

	/*
	ocs_deblevel=ODEB_CALL;
	*/

	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Binding\n"));
	
	sock = ocn_socket(AF_INET,SOCK_DGRAM,0);
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Bound to %d\n",sock));
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Closing %d\n",sock));
	ocn_close(sock);
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Binding 2\n"));
	sock = ocn_socket(AF_INET,SOCK_DGRAM,0);
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Bound to %d\n",sock));
	if ( sock >= 0 ) {
		err = ocn_bind(sock,&my_socket,sizeof(struct sockaddr_in));
		if ( err ){
			ODEBUGP(ODEB_CALL,( KERN_INFO "PING: bind error: %d\n",err));
			return(0);
		}
		else {
			ODEBUGP(ODEB_CALL,( KERN_INFO "PING: bind ok on: %d\n",sock));
		}
	} else {
		ODEBUGP(ODEB_CALL,( KERN_INFO "PING: socket error: %d\n",sock));
		return(0);
	}
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Bound to: %08x port %d\n",my_socket.sin_addr.s_addr, my_socket.sin_port));
	len = sizeof(struct sockaddr_in);
	if ( ocn_getname(sock, (struct sockaddr *)&my_socket , &len) ) {
		ODEBUGP(ODEB_CALL,( KERN_INFO "PING: ocn_getname error\n"));
		ocn_delete(sock);
		return(0);
	}
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: Bound to: %08x port %d\n",my_socket.sin_addr.s_addr, my_socket.sin_port));
	
	err=0;
	count2=0;
	while(err >= 0  && count2 < 100000 ){
		sprintf(my_message,"I shot the sheriff %04d",count2);
		my_message[24]=0;
		rtl_printf("PING: %s\n",my_message);
		err=ocn_sendto(sock,&my_message,MSG_LEN,MSG_WAITALL,&rem_socket,sizeof(struct sockaddr_in));
		if ( err >= 0 ) {
			count2++;
			len=sizeof(struct sockaddr_in);
			err = ocn_recvfrom(sock,my_receive_buffer,RECEIVE_BUFFER_SIZE,MSG_WAITALL,&rem_socket2,&len) ;
			my_receive_buffer[36]=0;
			rtl_printf("PING: %s\n",my_receive_buffer);
		}
	}
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: CALL CLOSE\n"));
	ocn_close(sock);
	ODEBUGP(ODEB_CALL,( KERN_INFO "PING: END\n"));
	return(0);
}
	

int init_module( void ) {
	int thread_status;
	pthread_attr_t attr;
	struct sched_param sched_param;

	pthread_attr_init(&attr);
	sched_param.sched_priority = 4;
	pthread_attr_setschedparam(&attr,&sched_param);
	thread_status=pthread_create(&ping_thread, &attr,ping_rt_thread,(void *)1); 

	if ( thread_status < 0 ){
		printk( KERN_INFO "ocera streams ping: pthread_create error.\n");
		return(-1);
		}

	printk( KERN_INFO "ocera ping (version "VERSION ") sucessfully loaded.\n"
        KERN_INFO "ocera ping: Copyright (C) 2003 Pierre Morel.\n" );
	return(0);
}


void cleanup_module( void )
{
	printk( KERN_INFO "ocera streams ping cleanup sock %d.\n",sock );
	ocn_delete(sock);
	pthread_delete_np (ping_thread);
	printk( KERN_INFO "ocera streams ping unloaded.\n" );
	return;
}

