/*
 * Pong: reply to UDP packet and increment a counter
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/socket.h>
#include <linux/in.h>

#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <asm/system.h>
#include <asm/io.h>

#include <rtl.h>
#include <rtl_fifo.h>

#define RTL_SPIN_LOCK SPIN_LOCK_UNLOCKED
#include <rtl_sync.h>
#include "../onetd.h"


MODULE_AUTHOR("Pierre Morel <pmorel@mnis.fr>");
MODULE_DESCRIPTION("Ocera Network pong module example");
MODULE_LICENSE("GPL");

/*
 * Initialize the module, get queues and wait for data
 */

int period = OCS_PERIOD*100 ;

pthread_t	pong_thread;


ocera_msgbuf_t	*tmp;

#define RECEIVE_BUFFER_SIZE	1024
char my_receive_buffer [RECEIVE_BUFFER_SIZE];


/* 192.168.254.5
#define	REM_ADDR	0xc0a8fe05
#define	REM_ADDR	htonl(INADDR_LOOPBACK)
*/
#define	LOC_ADDR	INADDR_ANY
#define MY_PORT	(unsigned short) 10061


static int sock=-1;
static struct sockaddr_in my_socket;
static struct sockaddr_in rem_socket;
char	*my_name = "pong";

/*
 * PONG THREAD
 */

void *pong_rt_thread(void *t){
int len;
int flags;
int err;

ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: pthread_make_periodic_np\n"));
pthread_make_periodic_np( pthread_self(), gethrtime(), period );
my_socket.sin_family = AF_INET;
my_socket.sin_addr.s_addr = LOC_ADDR;
my_socket.sin_port = htons(MY_PORT);


ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: Binding\n"));

sock=ocn_socket(AF_INET,SOCK_DGRAM,0);
if ( (err=ocn_bind(sock,&my_socket,sizeof(struct sockaddr_in)))){
	ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: Binding error %d\n",err));
	ocn_delete(sock);
	return(0);
	}
flags = MSG_WAITALL;
memset(my_receive_buffer,'A',20);
my_receive_buffer[20]=0;
while ( 1) {
	len = ocn_recvfrom(sock,my_receive_buffer,RECEIVE_BUFFER_SIZE,flags,&rem_socket,sizeof(struct sockaddr_in)) ;
	ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: Got %d data on socket %d\n",len,sock));
	ODEBUGP(ODEB_CALL,( KERN_INFO "PONG:***************************\n"));
	my_receive_buffer[25]=0;
	rtl_printf("PONG: %s\n",my_receive_buffer);
	ODEBUGP(ODEB_CALL,( KERN_INFO "PONG:***************************\n"));
	memset(my_receive_buffer,'A',20);
	}
ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: END\n"));
return(0);
}


int init_module( void )
{
	int thread_status;
	pthread_attr_t attr;
	struct sched_param sched_param;

	pthread_attr_init(&attr);
	sched_param.sched_priority = 4;
	pthread_attr_setschedparam(&attr,&sched_param);
	thread_status=pthread_create(&pong_thread, &attr,pong_rt_thread,(void *)1); 

if ( thread_status < 0 ){
	printk( KERN_INFO "ocera streams pong: pthread_create error.\n");
	return(-1);
	}

printk( KERN_INFO "ocera pong (version "VERSION ") sucessfully loaded.\n"
        KERN_INFO "ocera pong: Copyright (C) 2003 Pierre Morel.\n" );
return(0);
}


void cleanup_module( void )
{
	if(sock >= 0 ) {
		ocn_delete(sock);
	}
	pthread_delete_np (pong_thread);
	printk( KERN_INFO "ocera streams pong unloaded.\n" );
	return;
}

