/*
 * OCERA NETWORK INTERFACE
 *
 *
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/smp_lock.h>


#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <asm/system.h>
#include <asm/io.h>

#include "onetd.h"

struct proc_dir_entry *proc_ocn;
long ocera_net_state=0;

static int proc_calc_metrics(char *page, char **start, off_t off,
                 int count, int *eof, int len)
{
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;
    return len;
}


int proc_ocera_network_read_state(char *page, char **start, off_t off,
                 int count, int *eof, void *data)
{
int len;

len=sprintf(page,"%02lx\n",ocera_net_state);
return proc_calc_metrics(page, start, off, count, eof, len);
}

int ocn_read_proc_slot(char *page, char **start, off_t off,
                 int count, int *eof, void *data)
{
	int len;
	ocn_t	*sp = (ocn_t *)data;
	char	*def_name = "none";
	char	*u_name = def_name;
	char	*d_name = def_name;

	if ( sp->s_stream_up) { u_name = sp->s_stream_up->s_name; }
	if ( sp->s_stream_down) { d_name = sp->s_stream_down->s_name; }
	len=sprintf(page,"%02x %02x %04x %08x %05d %s %s\n",
					sp->s_sock,
					sp->s_type,
					sp->s_state,
					htonl(sp->s_sin.sin_addr.s_addr),
					htons(sp->s_sin.sin_port),
					u_name,
					d_name
					);
	return proc_calc_metrics(page, start, off, count, eof, len);
}



int ocn_socket_init(){
	int i;
	char name[20];
	struct proc_dir_entry *proc_slot;

	proc_ocn=proc_mkdir("onetd",proc_ocera);
	proc_slot=proc_mkdir("sockets",proc_ocn);

	for(i=0;i<OCERA_MAX_SOCKET;i++){
		memset(&ocera_socket[i], 0, sizeof(ocn_t));
		ocera_socket[i].s_state = OCN_SOCK_FREE;
		ocera_socket[i].s_sock = i ;
		sprintf(name,"socket_%02x",i);
		create_proc_read_entry(name, 0, proc_slot, ocn_read_proc_slot, &ocera_socket[i]);
	}
	return(0);
}

void ocn_reset_socket(ocn_t	*sp){
	memset(sp, 0, sizeof(ocn_t));
	sp->s_sock =  (sp - ocera_socket ) / sizeof(ocn_t);
	ODEBUGP(ODEB_CALL,("ocn_reset_socket: %x\n",sp->s_sock));
	sp->s_state = OCN_SOCK_FREE;
}

