#include <stdio.h>	/* printf, fprtinf */
#include <stdlib.h>	/* srand, rand */
#include <signal.h>

#if defined(WIN32)
#	include <windows.h>
#	include <windef.h>
#	include <ctype.h>
#	include <winbase.h>
#	include <winerror.h>
typedef SOCKET Socket;
#	define SOCKET_ERROR INVALID_SOCKET
#	define socket_close closesocket

#elif defined(_POSIX_C_SOURCE) || defined(__RTL__)
#	ifdef __RTL__
#		include <pthread.h>
#	endif
#	include <string.h>
#	include <unistd.h>
#	include <time.h>
#	include <sys/socket.h>	/* socket(), bind(), sendto(), recvfrom() */
#	include <netdb.h>		/* gethostbyname() */
#	include <arpa/inet.h>	/* inet_addr() */
typedef int Socket;
#	define SOCKET_ERROR -1
#ifdef __RTL__
#   ifndef ONETD
#	   define socket_close close_socket
#   else
#	   define socket_close ocn_close
#   endif

#else
#	define socket_close close
#endif

#else
#	error Target OS not supported !!!
#endif

#if defined(ONETD) && defined(__RTL__)
#include "onetd.h"
#endif

#define BASEPORT			5000
#define DATASIZE			256
#define MAX_SENT_PACKETS	50000
#define MAX_STOP_REQUESTS	10
#define STATS_PERIOD		(MAX_SENT_PACKETS/8)

#define BIND_ERROR      -2
#define HOST_ADDR_ERROR -3
#define SIZE_MAX_MESS  1024

int fin = 0;

Socket socket_create(struct sockaddr_in * my_addr_p, int my_port);
int resolve_host_addr(char* hostname, struct sockaddr_in * peer_addr_p, int port);
int server(Socket sock, struct sockaddr_in * peer_addr_p, int stats_period);
int client(Socket sock, struct sockaddr_in * peer_addr_p, int sendstop);
void termine(int e);

struct DataPacket
{
	int		session_id;
	int		packet_id;
	char	data[DATASIZE];
};
/* -------------------------------------------------------------------------- */
/*                             main                                           */
/* -------------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	Socket 				sock;
	int					i;
	int					my_port;
	struct sockaddr_in	my_addr;
	struct sockaddr_in	peer_addr;

#ifndef __RTL__
	/* Association du handler termine() au signal SIGINT (Ctrl-C) */
	signal(SIGINT, termine);
#endif

	/* Cration du socket */
	my_port = ( argc <= 1 ? BASEPORT : BASEPORT + 1 ); /* server.prot = BASEPORT, client.port = BASEPORT+1 */
	sock = socket_create(&my_addr, my_port);

	if (sock < 0)
		return sock;

	/* Connexion as server */
	if (argc <= 1)
	{
		int pr;
		printf("info: starting as a server; listening on port %d\n", BASEPORT);
		pr = server(sock, &peer_addr, STATS_PERIOD);
		printf("info: server exiting; total: %d packets received\n", pr);
	}
	/* Connexion as client */
	else
	{
		int ps;
		printf("info: starting as a client; connecting to host %s:%d\n", argv[1], BASEPORT);
		i = resolve_host_addr(argv[1], &peer_addr, BASEPORT);
		if (i < 0)
		{
			socket_close(sock);
			return i;
		}
		ps = client(sock, &peer_addr, (argc > 2) && (strcmp(argv[2], "-stopsrv") == 0));
		printf("info: client exiting; total: %d packets sent\n", ps);
	}

	socket_close(sock);

	return 0;
}

/* --------------------------------------------------------------- */
/*                       socket_create                             */
/* --------------------------------------------------------------- */

Socket socket_create(struct sockaddr_in * my_addr_p, int my_port)
{
	Socket sock;
	int err;



/*	printf("Opening socket...\n"); */
	#if defined(ONETD) && defined(__RTL__)
		sock = ocn_socket(AF_INET, SOCK_DGRAM, 0);
	#else
		sock = socket(AF_INET, SOCK_DGRAM, 0);
	#endif


	if (sock == SOCKET_ERROR)
	{
		fprintf(stderr, "error: unable to create udp socket\n");
		return SOCKET_ERROR;
	}

/*	printf("Binding socket... "); */
	memset (my_addr_p, 0, sizeof (struct sockaddr_in));
	my_addr_p->sin_family = AF_INET;
#if defined(WIN32)
	my_addr_p->sin_addr.S_un.S_addr = htonl (INADDR_ANY);
#elif defined(_POSIX_C_SOURCE) || defined(__RTL__)
	my_addr_p->sin_addr.s_addr = htonl(INADDR_ANY);
#endif
	my_addr_p->sin_port = htons (my_port);
	#if defined(ONETD) && defined(__RTL__)
		err = ocn_bind(sock, (struct sockaddr *)my_addr_p, sizeof(struct sockaddr_in));

	#else
		err = bind (sock, (struct sockaddr *)my_addr_p, sizeof(struct sockaddr_in));

	#endif

	if (err == -1)
	{
		fprintf(stderr, "error: unable to bind socket\n");
		socket_close(sock);
		#if defined(ONETD) && defined(__RTL__)
			ocn_close(sock);
		#else
			socket_close(sock);
		#endif


		return BIND_ERROR;
	}

	return sock;
}

/* --------------------------------------------------------------- */
/*                       resolve_host_addr                         */
/* --------------------------------------------------------------- */
int resolve_host_addr(char* hostname, struct sockaddr_in * peer_addr_p, int port)
{
	struct hostent		*host; /* note: sizeof(hostent) = 20 */
#ifndef ONETD // rajout de RP
	peer_addr_p->sin_family = AF_INET;
	peer_addr_p->sin_port = htons(port);
	host = gethostbyname(hostname);

#if defined(WIN32)
	if (host)
		peer_addr_p->sin_addr.S_un.S_addr = *((unsigned long *) host->h_addr_list[0]);
	else
		peer_addr_p->sin_addr.S_un.S_addr = inet_addr (hostname);
	if (peer_addr_p->sin_addr.S_un.S_addr == INADDR_NONE)
	{
		fprintf(stderr, "error: unable to get host server name\n");
		return HOST_ADDR_ERROR;
	}
#elif defined(_POSIX_C_SOURCE) || defined(__RTL__) 
	if (host)
		peer_addr_p->sin_addr.s_addr = *((unsigned long *) host->h_addr_list[0]);
	else
		peer_addr_p->sin_addr.s_addr = inet_addr (hostname);
	if (peer_addr_p->sin_addr.s_addr == INADDR_NONE)
	{
		fprintf(stderr, "error: unable to get host server name\n");
		return HOST_ADDR_ERROR;
	}
#endif
#endif  // rajout de RP

	return 0;
}
/*-------------------------------------------------------------------------*/
/*                             server                                      */
/*-------------------------------------------------------------------------*/

int server(Socket sock, struct sockaddr_in * peer_addr_p, int stats_period)
{
	int					i = 0, k = 0, lost = 0, session = 0;
	size_t		 		addr_len = sizeof(struct sockaddr_in);
	struct DataPacket	packet;
	size_t				packet_len;
	char message[10];
	size_t mess_len;
	int err;

#ifndef WIN32
	struct timespec		start, stop;
	unsigned int		dt;
#endif


#ifdef __RTL__
	printf("Calibration horloge (attente 1s)\n");
	clock_gettime(CLOCK_REALTIME, &start);
	usleep(1000000);
	clock_gettime(CLOCK_REALTIME, &stop);
	dt = (stop.tv_sec - start.tv_sec) * 1000000000 +
		(stop.tv_nsec - start.tv_nsec); /* nanoseconds */
	printf("Calibration = %d\n", dt);
#endif

	while (!fin)
	{
		/* Lecture d'un packet depuis le socket */
		#if defined(ONETD) && defined(__RTL__)
			packet_len = ocn_recvfrom(sock,
									  &packet, sizeof(struct DataPacket),
									  0,
									  (struct sockaddr*)peer_addr_p, &addr_len
									  );

		#else
			packet_len = recvfrom(sock, 
								  &packet, sizeof(struct DataPacket), 
								  0, 
								  (struct sockaddr*)peer_addr_p, &addr_len
								  );

		#endif


		/* Test for MESG request from client */
		if (strncmp(packet.data, "MESG", 4) == 0)
		{
			printf("info: server: message from client:\n%s\n", packet.data);
			continue;
		}

		/* Test for STOP request from client */
		if (strncmp(packet.data, "STOP", 4) == 0)
		{
			printf("info: server: exit requested by client; exiting\n");
			break;
		}

		/* Check for session id --> reset state if new session */
		if (session != packet.session_id)
		{
			session = packet.session_id;
			i = 0;
			k = 0;
			lost = 0;
			printf("info: server: new session id %d started\n", session);
		}

		/* How many packets have been lost ? */
		lost += packet.packet_id - i;
		i = packet.packet_id + 1;

		if (k % stats_period == 0)
		{
			if (k == 0)
				clock_gettime(CLOCK_REALTIME, &start);
			else
			{
				unsigned int pps;
				clock_gettime(CLOCK_REALTIME, &stop);
/*				dt = (stop.tv_sec - start.tv_sec) +
				     (stop.tv_nsec - start.tv_nsec) * 1E-9;
*/				dt = (stop.tv_sec - start.tv_sec) * 1000000000 +
				     (stop.tv_nsec - start.tv_nsec); /* nanoseconds */
				pps = 10000*stats_period/(dt/100000); /* pckts/s */
				printf("%d packets received in %d ms; rate = %d pckts/s = %d KB/s; %d packet(s) lost (%d%%)\n",
						stats_period,
						dt/1000000,
						pps,
						pps*sizeof(struct DataPacket)/1024,
						lost,
						100*lost/stats_period);
				start = stop;
				lost = 0;
			}
		}
		k++;

		strcpy(message, "Recepisse12"); /* taille=9+1 */
		mess_len = strlen(message) + 1;

/*		printf("sendto(id socket=%d, message=^0x%x, message length=%d, 0 pour flags, "
		"destinataire=^0x%x, );\n",
		sock, message, mess_len,  peer_addr_p); */

		#if defined(ONETD) && defined(__RTL__)
			err = ocn_sendto(sock, 
				  (void *)message, mess_len, 
				  0,
				  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
		#else
			err = sendto(sock, 
				  (void *)message, mess_len, 
				  0,
				  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));

		#endif

		
	}
	return k;
}

/*------------------------------------------------------------------------*/
/*                            client                                      */
/*------------------------------------------------------------------------*/

int client(Socket sock, struct sockaddr_in * peer_addr_p, int sendstop)
{
	int	i;
	struct DataPacket	packet;
	size_t		 		addr_len = sizeof(struct sockaddr_in);
	size_t				packet_len;
	//void *messageAck;
	char messageAck[SIZE_MAX_MESS];
	int err;

#ifndef WIN32
	struct timespec		time;
#endif

	/* randomize session id */
#ifndef WIN32
	clock_gettime(CLOCK_REALTIME, &time);
	srand(time.tv_nsec);
#endif
	packet.session_id = rand();

	printf("info: client: sending welcome message to server\n");
	packet.packet_id = 0;
	strcpy(packet.data, "MESG: client tells HELLO to server");
	#if defined(ONETD) && defined(__RTL__)
		err = ocn_sendto(sock, 
			  (void *)&packet, sizeof(struct DataPacket), 
			  0,
			  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
	printf("udptest info: client: retour ocn_sendto\n");

	#else
		err = sendto(sock, 
			  (void *)&packet, sizeof(struct DataPacket), 
			  0,
			  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
	#endif

	i = 0;
	while (!fin && i<=MAX_SENT_PACKETS)
	{
	printf("udptest info: client: on est dans la boucle while\n");
		/* Fill packet with data */
		packet.packet_id = i;
		strcpy(packet.data, "DATA: This is some data to be sent...");

		/* Emmission d'un packet vers le serveur hote */

		#if defined(ONETD) && defined(__RTL__)
			err = ocn_sendto(sock, 
				  (void *)&packet, sizeof(struct DataPacket), 
				  0,
				  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
		#else
			err = sendto(sock, 
				  (void *)&packet, sizeof(struct DataPacket), 
				  0,
				  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
		#endif
		
/* DEBUG
#ifndef __RTL__
sleep(3);
#endif*/

		if (i % STATS_PERIOD == 0)
			printf("%d packets sent; session_id=%d, packet_id=%d, data=%s\n",
				    STATS_PERIOD,
					packet.session_id, packet.packet_id, packet.data);

		i ++;

		/* Lecture d'un packet depuis le socket */
/*		printf("attente rception acknoledge serveur\n"); */

		/*ssize_t recvfrom(int socket, void *buffer, size_t length, int flags,
			struct sockaddr *address, socklen_t *address_len)*/

		#if defined(ONETD) && defined(__RTL__)
			packet_len = ocn_recvfrom(sock,
									  (void *)messageAck, SIZE_MAX_MESS,
									  0,
									  (struct sockaddr*)peer_addr_p, &addr_len);
		#else
			packet_len = recvfrom(sock, 
								  (void *)messageAck, SIZE_MAX_MESS, 
								  0, 
								  (struct sockaddr*)peer_addr_p, &addr_len);
		#endif
/*		printf("Reception message: %s (taille=%d)\n", messageAck, packet_len); */
	}

	if (sendstop)
	{
		int j;
		printf("info: client: sending %d STOP requests to server\n", MAX_STOP_REQUESTS);
		packet.packet_id = i;
		strcpy(packet.data, "STOP");
		for (j = 0; j < MAX_STOP_REQUESTS; j++)
			{
			#if defined(ONETD) && defined(__RTL__)
				err = ocn_sendto(sock, 
					  (void *)&packet, sizeof(struct DataPacket), 
					  0,
					  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
			#else
				err = sendto(sock, 
					  (void *)&packet, sizeof(struct DataPacket), 
					  0,
					  (struct sockaddr *) peer_addr_p, sizeof(struct sockaddr_in));
			#endif	
			}
	
	}

	return i;
}

void termine(int e)
{
	fin = 1;
}

#ifdef __RTL__

pthread_t task = NULL;

void* task_code(void* arg)
{
	return (void*) main(0, NULL); /* server mode */
}

int init_module (void)
{
	pthread_attr_t attr;
	struct sched_param sched_param;

	pthread_attr_init(&attr);
	sched_param.sched_priority = sched_get_priority_max(SCHED_FIFO) - 10;
	pthread_attr_setschedparam(&attr, &sched_param);
	pthread_attr_setfp_np(&attr, 1);
	pthread_create(&task, &attr, task_code, (void *) NULL);

	return 0;
}

void cleanup_module (void)
{
	pthread_cancel(task);
	pthread_join(task, NULL);
}

#endif

#ifdef RTSCKT_TIMED_SELECT
#	include <sys/time.h>
/* int timed_select(int fd, int timeout) : select on fd with timeout
   fd:      file-descriptor
   timeout: timeout in usec */
int timed_select(int fd, int timeout)
{
	fd_set			rfds;
	struct timeval	tv;
	int				s;

	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	tv.tv_sec = 0;
	tv.tv_usec = timeout;
	s = -1;
	while ((s != 0) && (!FD_ISSET(fd, &rfds)))
	{
		s = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
	}
	return s;
}
#endif


