#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/socket.h>
#include <linux/in.h>

#include <pthread.h>
#include <time.h>
#include <unistd.h>

#include <asm/system.h>
#include <asm/io.h>

#include <rtl.h>
#include <rtl_fifo.h>

#define RTL_SPIN_LOCK SPIN_LOCK_UNLOCKED
#include <rtl_sync.h>
#include "../ocs.h"


MODULE_AUTHOR("Pierre Morel <pmorel@mnis.fr>");
MODULE_DESCRIPTION("Ocera Network pong module example");
MODULE_LICENSE("GPL");

/*
 * Initialize the module, get queues and wait for data
 */

int period = OCS_PERIOD*100 ;

pthread_t	pong_thread;
char * mon_message = "salut";
ocs_t	* my_stream_p;


void *pong_rt_thread(void *t){
	ocs_t	*sp;
	ocera_msgbuf_t *mbp;

	ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: start %s\n",t));

	ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: pthread_make_periodic_np\n"));
	pthread_make_periodic_np( pthread_self(), gethrtime(), period );

	if ( (sp=ocs_register("pong",ST_RTLINUX_TYPE)) == NULL) {
		ODEBUGP(ODEB_CALL,( KERN_INFO "PONG: ocs_register error\n"));
		return(0);
	}
	my_stream_p = sp ;
	do {
		ODEBUGP(ODEB_CALL,("PONG: loop sp = %p\n",sp));
		if( ( mbp = ocs_get_q(&sp->s_queue_up, OCS_WAIT)) == NULL) {
			ODEBUGP(ODEB_CALL,("PONG: ocs_get_q error\n"));
			return(0);
		}
		ocs_put_q(&sp->s_queue_down,mbp);
		ODEBUGP(ODEB_CALL,("PONG: wait next period\n"));
		pthread_wait_np();
		} while(1);
	return(0);
}


int init_module( void )
{
	int thread_status;
	pthread_attr_t attr;
	struct sched_param sched_param;

	pthread_attr_init(&attr);
	sched_param.sched_priority = 4;
	pthread_attr_setschedparam(&attr,&sched_param);
	thread_status=pthread_create(&pong_thread, &attr,pong_rt_thread,(void *)mon_message); 

if ( thread_status < 0 ){
	printk( KERN_INFO "ocera streams pong: pthread_create error.\n");
	return(-1);
	}

printk( KERN_INFO "ocera pong (version "VERSION ") sucessfully loaded.\n"
        KERN_INFO "ocera pong: Copyright (C) 2003 Pierre Morel.\n" );
return(0);
}


void cleanup_module( void )
{
ocs_unregister(my_stream_p);
pthread_delete_np (pong_thread);
printk( KERN_INFO "ocera streams pong unloaded.\n" );
return;
}

