/*
 * OCERA STREAMS INCLUDE
 *
 *
 */
#ifndef OCS
#define OCS

#include <semaphore.h>

#ifndef NULL
#define NULL 0
#endif

#define MAX_OCERA_STREAMS 64

#define ODEB_NONE		0
#define ODEB_ERROR		1
#define ODEB_DEB		2
#define ODEB_CALL		3
#define ODEB_STREAM		4
#define ODEB_FULL		5
#define ODEB_ALL		5

#define     OCS_PERIOD  (HRTICKS_PER_SEC/HZ)/2


#define ODEBUG	1
#ifdef ODEBUG
extern int	ocs_deblevel;
#define ODEBUGP(level,args) { if(ocs_deblevel >= level ) rtl_printf args; }
#else
#define ODEBUGP(level,args) {}
#endif

typedef struct _ocs_t    ocs_t;

typedef struct _ocera_msgbuf_t    ocera_msgbuf_t;

typedef struct _msg_q       ocera_msgbuf_q;

#define NULLMSGP	(ocera_msgbuf_t *) 0
#define MAXMSGSZ    32

struct _ocera_msgbuf_t {
    ocera_msgbuf_t      *mb_next;
    ocera_msgbuf_t		*mb_last;
    int					mb_count;
    int					mb_flags;
	void				*mb_datap;
	int					mb_type;
    unsigned char       mb_data[MAXMSGSZ];
    };

#define MAX_MSGBUF  128

struct _msg_q   {
    ocera_msgbuf_q  	*q_next;
    ocera_msgbuf_t		*q_head;
    ocera_msgbuf_t		*q_tail;
    unsigned int		q_count;
    unsigned int		q_status;
	int					q_soft;
	spinlock_t			q_lock;
	wait_queue_head_t	q_sem_s;
	sem_t				q_sem_r;
	ocs_t				*q_stream;
    };


#define	ST_MODULE_TYPE	1
#define	ST_DRIVER_TYPE	2
#define	ST_HEADER_TYPE	4
#define	ST_LINUX_TYPE	8
#define	ST_RTLINUX_TYPE	16

#define ST_MAX_NAME_LEN	16

struct _ocs_t   {
	long			s_state;
	long			s_type;
	char 			s_name[ST_MAX_NAME_LEN];
	ocera_msgbuf_q	s_queue_up;
	ocera_msgbuf_q	s_queue_down;
	ocera_msgbuf_q	s_queue_wait;
	ocs_t	*		s_up;
	ocs_t	*		s_down;
    };

extern ocs_t         	ocs_table[];
extern ocera_msgbuf_q	mbp_free_q;

#define	OCS_NOWAIT		0x01
#define	OCS_WAIT		0x02
#define	OCS_RTL			0x04

extern int				ocs_sleep(ocera_msgbuf_q *, int );
extern void				ocs_wakeup(ocera_msgbuf_q *);

extern ocera_msgbuf_t   *ocs_get_q(ocera_msgbuf_q *, int);
extern void    			ocs_put_q(ocera_msgbuf_q *,ocera_msgbuf_t *);
extern void    			ocs_put_nextq(ocs_t *,ocera_msgbuf_t *);
extern void				ocs_free_q(ocera_msgbuf_q *);

extern void				ocs_init(void);
extern ocs_t *			ocs_register(char *,int );
extern void				ocs_unregister(ocs_t *);

extern void				ocs_bind(ocs_t * , ocs_t *);
extern void				ocs_unbind(ocs_t *);

extern ocera_msgbuf_t *	ocs_get_mbuf(int);
extern void				ocs_free_mbuf(ocera_msgbuf_t *);


#define OCERA_STREAMS_ISR	1

#define OCERA_STREAMS_STATE_VOID	0
#define OCERA_STREAMS_STATE_INIT	1
#define OCERA_STREAMS_STATE_BUSY	2
#define OCERA_STREAMS_STATE_FREE	4
#define OCERA_STREAMS_STATE_RUNNING	8

extern long ocs_state;

extern struct proc_dir_entry *proc_ocera;
extern struct proc_dir_entry *proc_ocs;
extern struct proc_dir_entry *proc_ocs_slot;

extern int ocs_read_proc_state(char *page, char **start, off_t off, int count, int *eof, void *data);
extern int ocs_read_proc_msgbuf(char *page, char **start, off_t off, int count, int *eof, void *data);

extern int ocs_read_proc_slot(char *page, char **start, off_t off, int count, int *eof, void *data);
extern int ocs_deblevel_read(char * buffer, char ** start, off_t offset, int size, int * eof, void * data);
extern int ocs_deblevel_write(struct file * filp, const char * buffer, unsigned long count, void * data);
extern void ocs_release_all(void);

extern int ocs_free;

#define OCS_EXTERNAL_BUFFER	0
#define OCS_STREAM_BUFFER	1

#endif
