/*
 * OCERA STREAMS
 *
 *
 */
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/proc_fs.h>

#include <pthread.h>
#include <time.h>
#include <unistd.h>
#include <linux/smp_lock.h>

#include <asm/system.h>
#include <asm/io.h>


#define RTL_SPIN_LOCK SPIN_LOCK_UNLOCKED
#include <rtl.h>
#include <rtl_printf.h>
#include <rtl_sync.h>
#include "ocs.h"

MODULE_AUTHOR("Pierre Morel <pmorel@mnis.fr>");
MODULE_DESCRIPTION("Ocera Streams");
MODULE_LICENSE("GPL");

int  ocs_deblevel=ODEB_NONE;
MODULE_PARM(ocs_deblevel,"i");
MODULE_PARM_DESC(ocs_deblevel, "The DEBUG LEVEL: 0/1/2/3 for none/error/deb/call");

extern void ocs_freequeue_isr(int , void *, struct pt_regs *);
extern void ocs_isr(int , void *, struct pt_regs *);


spinlock_t	ocs_table_lock = SPIN_LOCK_UNLOCKED ;
ocs_t		ocs_table[MAX_OCERA_STREAMS];
ocs_t		*ocs_current;

ocera_msgbuf_t        ocera_msgbufs[MAX_MSGBUF];

unsigned int    num_msg = MAX_MSGBUF;

ocera_msgbuf_q        mbp_free_q;



long ocs_state = OCERA_STREAMS_STATE_VOID;
static char *default_name = "FREE" ;
int		ocs_free = MAX_OCERA_STREAMS;


/*
 * Initialization routine
 */

int init_module(void)
{
	register short       i=0;
	register ocs_t *ost_p;
	char name[5];
	struct proc_dir_entry * pdep;

	printk( KERN_INFO  "Ocera Streams ("VERSION") ");
	if( ocs_state != OCERA_STREAMS_STATE_VOID ) {
		return(-1);
		}
	ocs_state = OCERA_STREAMS_STATE_INIT ;

	proc_ocera=proc_mkdir("ocera",0);
	proc_ocs=proc_mkdir("ocs",proc_ocera);
	proc_ocs_slot=proc_mkdir("streams_slot",proc_ocs);
	create_proc_read_entry("state", 0, proc_ocs, ocs_read_proc_state, NULL);
	create_proc_read_entry("freemsg", 0, proc_ocs, ocs_read_proc_msgbuf, NULL);
	pdep = create_proc_entry("debug", S_IRUGO | S_IWUSR, proc_ocs);
	if (pdep) {
		pdep->write_proc = ocs_deblevel_write;
		pdep->read_proc = ocs_deblevel_read;
	}


	printk( "(C)2004 Pierre Morel - MNIS\n");
/* initialise the streams tables                  */
	ocs_table_lock=SPIN_LOCK_UNLOCKED;
    for(i=0;i<MAX_OCERA_STREAMS;i++) {
		ost_p = &ocs_table[i];
		memset(ost_p,0,sizeof(ocs_t));
        ost_p->s_state = OCERA_STREAMS_STATE_FREE;
        strncpy(ost_p->s_name,default_name,strlen(default_name));
        ost_p->s_queue_up.q_stream=ost_p;
        ost_p->s_queue_down.q_stream=ost_p;
        ost_p->s_queue_wait.q_stream=ost_p;
		ost_p->s_queue_up.q_lock=SPIN_LOCK_UNLOCKED;
		ost_p->s_queue_down.q_lock=SPIN_LOCK_UNLOCKED;
		ost_p->s_queue_wait.q_lock=SPIN_LOCK_UNLOCKED;

		sprintf(name,"%02x",i);
		create_proc_read_entry(name, 0, proc_ocs_slot, ocs_read_proc_slot, &ocs_table[i]);
        }

/* initialise the messages buffers                  */
    for(i=1;i<num_msg-1;i++) {
        ocera_msgbufs[i].mb_next = &ocera_msgbufs[i+1];
        ocera_msgbufs[i].mb_last = &ocera_msgbufs[i-1];
        }
    ocera_msgbufs[num_msg-1].mb_next = NULL;
    ocera_msgbufs[num_msg-1].mb_last = &ocera_msgbufs[i-2];
    ocera_msgbufs[0].mb_next = &ocera_msgbufs[1];
    ocera_msgbufs[0].mb_last = NULL;
    mbp_free_q.q_head = &ocera_msgbufs[0];
    mbp_free_q.q_tail = &ocera_msgbufs[num_msg-1];
    mbp_free_q.q_count= num_msg;
	mbp_free_q.q_soft = rtl_get_soft_irq (ocs_freequeue_isr, "mbp_free_q");
	if ( mbp_free_q.q_soft <= 0 ){
		printk( KERN_INFO "cannot allocate mbp_free_q irq! ABORT.\n");
		return(-1);
	}
	ocs_state = OCERA_STREAMS_STATE_RUNNING ;

	printk( "Ocera Streams sucessfully loaded.\n");

return(0);
}

void cleanup_module( void )
{
	if ( ocs_state == OCERA_STREAMS_STATE_RUNNING ) {
		ocs_state = OCERA_STREAMS_STATE_BUSY;
		ocs_release_all();
		rtl_free_soft_irq(mbp_free_q.q_soft);
		remove_proc_entry("streams_slot",proc_ocs);
		remove_proc_entry("ocs",proc_ocera);
		remove_proc_entry("ocera",0);
	}
	printk( KERN_INFO "Ocera Streams unloaded.\n" );
	return;
}




/************************************************************************
*	FUNCTION:	ocs_isr
*
*	DESCRIPTION:
*		called by the interrupt handler mecanism for soft_irq.
*		polls the streams to see if any queuehas the STATE_ISR
*		bit set.
*
*************************************************************************/


void ocs_freequeue_isr(int irq, void *dev_id, struct pt_regs *p){
	ODEBUGP(ODEB_STREAM,("ost: ocs_freequeue_isr: %d - %p \n",irq,dev_id));
	wake_up_interruptible(&mbp_free_q.q_sem_s);
	current->need_resched = 1;
}

void ocs_isr(int irq, void *dev_id, struct pt_regs *p){
	ocs_t *s;
	int i;

	ODEBUGP(ODEB_STREAM,("ost: ocs_isr: %d - %p \n",irq,dev_id));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) return;

	for( i=0; i< MAX_OCERA_STREAMS ; i++ ) {
		s = &ocs_table[i];
		if ( s->s_queue_up.q_soft == irq ) {
			if ( s->s_state == OCERA_STREAMS_STATE_BUSY){
				ODEBUGP(ODEB_STREAM,("ost: ocs_isr: wake irq: %d for stream %d on queue: %p \n",irq,i,&s->s_queue_up));
				wake_up_interruptible(&s->s_queue_up.q_sem_s);
				wake_up_interruptible(&s->s_queue_down.q_sem_s);
				wake_up_interruptible(&s->s_queue_wait.q_sem_s);
				current->need_resched = 1;
				return;
			}
		}
	}
	ODEBUGP(ODEB_STREAM,("ost: ocs_isr: Spurious interrupt %d\n",irq));
}




/************************************************************************
**	FUNCTION:	ocs_register
*************************************************************************/

ocs_t * ocs_register(char * name , int type )
{
	register ocs_t *ost_p;
	register	int	i;
	unsigned long		flags;

	ODEBUGP(ODEB_STREAM,("ost: ocs_register name: %s - type %d\n",name, type));

	rtl_spin_lock_irqsave(&ocs_table_lock,flags);
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
		return((ocs_t *)0);
		}


	for( i=0; i< MAX_OCERA_STREAMS ; i++) {
		ost_p = &ocs_table[i];
		if( ost_p->s_state == OCERA_STREAMS_STATE_FREE ) {
			ost_p->s_state=OCERA_STREAMS_STATE_BUSY;
        	strncpy(ost_p->s_name,name,strlen(name));
			ost_p->s_type=type;
			ocs_free--;
			if ((type & ST_LINUX_TYPE )) {
				if	((ost_p->s_queue_up.q_soft = rtl_get_soft_irq (ocs_isr, name)) == 0 ){
					ost_p->s_state=OCERA_STREAMS_STATE_FREE;
					ocs_free++;
					ODEBUGP(ODEB_STREAM,("ost: Cannot allocate SoftIRQ\n"));
					ODEBUGP(ODEB_STREAM,("ost: ocs_register END : 0\n"));
					rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
					return((ocs_t *) 0);
				}
				ODEBUGP(ODEB_STREAM,("ost: SoftIRQ: %d\n",ost_p->s_queue_up.q_soft));
				ost_p->s_queue_down.q_soft = ost_p->s_queue_up.q_soft;
				ost_p->s_queue_wait.q_soft = ost_p->s_queue_up.q_soft;
				init_waitqueue_head(&ost_p->s_queue_up.q_sem_s);
				init_waitqueue_head(&ost_p->s_queue_down.q_sem_s);
				init_waitqueue_head(&ost_p->s_queue_wait.q_sem_s);
			} else {
				ost_p->s_queue_up.q_soft = 0 ;
				sem_init (&ost_p->s_queue_down.q_sem_r, 1, 0);
				sem_init (&ost_p->s_queue_up.q_sem_r, 1, 0);
				sem_init (&ost_p->s_queue_wait.q_sem_r, 1, 0);
			}
			ODEBUGP(ODEB_STREAM,("ost: ocs_register END : stream: %d at %p\n",i,ost_p));
			rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
			return(ost_p);
			}
		}
	rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
	ODEBUGP(ODEB_STREAM,("ost: ocs_register END : 0\n"));
	return((ocs_t *) 0);
}

/************************************************************************
**	FUNCTION:	ocs_bind
*************************************************************************/
void ocs_bind(ocs_t* up , ocs_t *down)
{
	unsigned long flags;

	ODEBUGP(ODEB_STREAM,("ost: ocs_bind %x - %x\n",up, down));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		return;
		}
	rtl_spin_lock_irqsave(&ocs_table_lock,flags);
	up->s_down = down;
	down->s_up = up;
	up->s_queue_down.q_next = &down->s_queue_down ;
	down->s_queue_up.q_next = &up->s_queue_up ;
	rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
	ODEBUGP(ODEB_STREAM,("ost: ocs_bind END\n"));
}

/************************************************************************
**	FUNCTION:	ocs_unbind
*************************************************************************/
void ocs_unbind(ocs_t* s )
{
	unsigned long flags;
	ODEBUGP(ODEB_STREAM,("ost: ocs_unbind %x\n",s));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) return;
	rtl_spin_lock_irqsave(&ocs_table_lock,flags);
	if ( s->s_down ) {
		s->s_down->s_up = 0;
		s->s_down->s_queue_up.q_next = 0;
		s->s_down=0;
	}
	if ( s->s_up ) {
		s->s_up->s_down = 0;
		s->s_up->s_queue_down.q_next = 0;
		s->s_up=0;
	}
	rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
	ODEBUGP(ODEB_STREAM,("ost: ocs_unbind END\n"));
}

/************************************************************************
**	FUNCTION:	ocs_unregister
*************************************************************************/
void ocs_unregister(ocs_t *ost_p)
{
	unsigned long flags;

	ODEBUGP(ODEB_STREAM,("ost: ocs_unregister %x free: %d\n",ost_p,ocs_free));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		ODEBUGP(ODEB_STREAM,("ost: ocs_unregister NOT RUNNING\n"));
		return;
		}
	ODEBUGP(ODEB_STREAM,("ost: ocs_unregister %s state %d\n",ost_p->s_name,ost_p->s_state));
	rtl_spin_lock_irqsave(&ocs_table_lock,flags);
	if ( ost_p->s_state != OCERA_STREAMS_STATE_BUSY ){
		rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
		ODEBUGP(ODEB_STREAM,("ost: ocs_unregister allready free\n"));
		return;
	}
	ocs_free_q(&ost_p->s_queue_up);
	ocs_free_q(&ost_p->s_queue_down);
	ocs_free_q(&ost_p->s_queue_wait);
	memset(ost_p,0,sizeof(ocs_t));
	ost_p->s_state=OCERA_STREAMS_STATE_VOID;
    strncpy(ost_p->s_name,default_name,strlen(default_name));
	rtl_spin_unlock_irqrestore(&ocs_table_lock,flags);
	if ( ost_p->s_queue_up.q_soft) {
		ODEBUGP(ODEB_STREAM,("ost: ocs_unregister free soft IRQ %d\n",ost_p->s_queue_up.q_soft));
		rtl_free_soft_irq(ost_p->s_queue_up.q_soft);
	} else {
		sem_destroy(&ost_p->s_queue_up.q_sem_r);
		sem_destroy(&ost_p->s_queue_down.q_sem_r);
	}
	ocs_free++;
	ost_p->s_state=OCERA_STREAMS_STATE_FREE;
	ODEBUGP(ODEB_STREAM,("ost: ocs_unregister END: free: %d\n",ocs_free));
	return;
}

/************************************************************************
**							 UTILITIES    
*************************************************************************/


/************************************************************************
**      FUNCTION : ocs_free_mbuf()               
************************************************************************/

void ocs_free_mbuf(mbp)
ocera_msgbuf_t *mbp;
{
	ODEBUGP(ODEB_ALL,("ost: ocs_free_mbuf %x\n",mbp));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		return;
	}
   	if(mbp == NULL)
       	return;
    ocs_put_q(&mbp_free_q,mbp);
}




/************************************************************************
**      FUNCTION : ocs_get_mbuf()                 
************************************************************************/


ocera_msgbuf_t * ocs_get_mbuf(int mode)
{
	ocera_msgbuf_t *mbp;

	ODEBUGP(ODEB_ALL,("ost: ocs_get_mbuf %d\n",mode));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		return(NULL);
	}
	if( ( mbp = ocs_get_q(&mbp_free_q, mode)) == NULL) {
    	return(NULL);
    	}
	memset(mbp, 0, sizeof(ocera_msgbuf_t));
	return(mbp);
}

/************************************************************************
**      FUNCTION    ocs_free_q()           
************************************************************************/
void ocs_free_q(ocera_msgbuf_q *q){
	ocera_msgbuf_t  *mbp;
	while( (mbp = ocs_get_q(q,OCS_NOWAIT)) != NULL) {
		ODEBUGP(ODEB_STREAM,("ost: ocs_free_q freing %x on %x\n",mbp,q));
		ocs_free_mbuf(mbp);
	}
}
/************************************************************************
**      FUNCTION    ocs_get_q()           
************************************************************************/

ocera_msgbuf_t    * ocs_get_q(q, mode)
ocera_msgbuf_q		*q;
int					mode;
{
	unsigned long	flags;
	ocera_msgbuf_t	*tmp;
	int				ret;

	ODEBUGP(ODEB_STREAM,("ost: ocs_get_q 0x%x (%d)\n",q,q->q_count));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		return(NULL);
	}
	rtl_spin_lock_irqsave(&q->q_lock,flags);
	while ( q->q_count <= 0 ) {
		if ( mode & OCS_NOWAIT)  {
			rtl_spin_unlock_irqrestore(&q->q_lock,flags);
			ODEBUGP(ODEB_STREAM,("ost: ocs_get_q END1\n",q));
			return(NULL);
		}
		rtl_spin_unlock_irqrestore(&q->q_lock,flags);
		ODEBUGP(ODEB_STREAM,("ost: ocs_get_q goes to sleep\n",q));
		if ( (ret = ocs_sleep(q,mode)) != 0 ) {
			ODEBUGP(ODEB_STREAM,("ost: ocs_get_q signals up\n",q));
			return(NULL);
		}
		ODEBUGP(ODEB_STREAM,("ost: ocs_get_q wakes up\n",q));
		rtl_spin_lock_irqsave(&q->q_lock,flags);
    }
	tmp=q->q_head;
	q->q_head = tmp->mb_next;
	q->q_count--;
	if(q->q_count == 0){
    	q->q_head = NULL;
    	q->q_tail = NULL;
    	}
	tmp->mb_next = NULL;
	tmp->mb_last = NULL;
	rtl_spin_unlock_irqrestore(&q->q_lock,flags);
	ODEBUGP(ODEB_STREAM,("ost: ocs_get_q got %x END\n",tmp));
	return(tmp);
}
/************************************************************************
**	FUNCTION:	ocs_put_nextq
*************************************************************************/

void ocs_put_netxq(ocs_t *s,ocera_msgbuf_t *mbp)
{
	ODEBUGP(ODEB_ALL,("ost: ocs_put_netxq %x %x\n",s,mbp));
	ocs_put_q(&s->s_up->s_queue_up,mbp);
}

/************************************************************************
**	FUNCTION:	ocs_put_q
*************************************************************************/

void ocs_put_q(q,mbp)
ocera_msgbuf_q    *q;
ocera_msgbuf_t    *mbp;
{
	unsigned long flags;

	ODEBUGP(ODEB_STREAM,("ost: ocs_put_q %x %x\n",q,mbp));
	if ( ocs_state != OCERA_STREAMS_STATE_RUNNING ) {
		return;
	}
	rtl_spin_lock_irqsave(&q->q_lock,flags);
	mbp->mb_next = NULL;
	if(q->q_head == NULL ){
    	q->q_head = mbp;
    	q->q_tail = mbp;
    	q->q_count = 1;
		rtl_spin_unlock_irqrestore(&q->q_lock,flags);
		ocs_wakeup(q);
    	}
	else    {
    	q->q_tail->mb_next = mbp;
    	q->q_tail = mbp;
    	q->q_count++;
		rtl_spin_unlock_irqrestore(&q->q_lock,flags);
    	}
	ODEBUGP(ODEB_STREAM,("ost: ocs_put_q END\n"));
}

/*
 * ocs_sleep
 */
int	ocs_sleep(ocera_msgbuf_q *q, int mode)
{
	int ret=0;
	ODEBUGP(ODEB_STREAM,("ost: ocs_sleep on queue %p\n",q));

	if ( mode & OCS_RTL) {
		ODEBUGP(ODEB_STREAM,("ost: ocs_sleep hard on sem %p\n",&q->q_sem_r));
		ret = sem_wait(&q->q_sem_r);
	}
	else if ( q->q_soft ) {
		ODEBUGP(ODEB_STREAM,("ost: ocs_sleep soft on sem %p\n",&q->q_sem_s));
		interruptible_sleep_on(&q->q_sem_s);
		ret=signal_pending(current);
	} else {
		ODEBUGP(ODEB_STREAM,("ost: ocs_sleep: BAD CALL END\n",ret));
		ret=-1;
	}
	ODEBUGP(ODEB_STREAM,("ost: ocs_sleep: %d END\n",ret));
	return ret ;
}
/*
 * ocs_wakeup
 */

void	ocs_wakeup(q)
ocera_msgbuf_q    *q;
{
	ODEBUGP(ODEB_STREAM,("ost: ocs_wakeup: q = %p\n",q));
	if ( q->q_soft ){
		ODEBUGP(ODEB_STREAM,("ost: ocs_wakeup soft: %p irq: %d\n",&q->q_sem_s,q->q_soft));
		rtl_global_pend_irq(q->q_soft);
	} else {
		ODEBUGP(ODEB_STREAM,("ost: ocs_wakeup hard: %p\n",&q->q_sem_r));
		sem_post(&q->q_sem_r);
		}
	ODEBUGP(ODEB_STREAM,("ost: ocs_wakeup END\n"));
}

/************************************************************************
**	FUNCTION:	ocs_release_all
*************************************************************************/

void ocs_release_all(){
	ocs_t	*ost_p;
	int		i;

	for( i=0; i< MAX_OCERA_STREAMS ; i++) {
		ost_p = &ocs_table[i];
		if( ost_p->s_state == OCERA_STREAMS_STATE_BUSY ) {
			ocs_unregister(ost_p);
		}
	}
}
